#ifndef BOUNDING_BOX_INCLUDED
#define BOUNDING_BOX_INCLUDED
#include "geometry.h"

/* This class represents a bounding box object that surrounds a shape.
 * The box is assumed to be lined up with the coordinate axis and is 
 * defined by (any) two antipodal points on the box. The constructors
 * make sure that the the (x,y,z) coordinates of p[0] are all less than
 * or equal to the (x,y,z) coordinates of p[1] */

class BoundingBox {
  void setBoundingBox(Point3D p1,Point3D p2);
 public:
  Point3D p[2];

  /* This constructor creates a bounding box representing the point (0,0,0) */
  BoundingBox(void);
  /* This constructor creates a bounding box with the specified antipodal 
   * vertices */
  BoundingBox(Point3D p1,Point3D p2);

  /* Given a list of points, this constructor creates the smallest bounding
   * box containing all of the points */
  BoundingBox(Point3D* pList,int pSize);

  /* This method returns the bounding box containing the union of the two
   * bounding boxes */
  BoundingBox operator+ (BoundingBox b);

  /* This method returns the bounding box generated by first transforming the
   * initial bounding box according to the specified transformation and then
   * finding the minimal axis-aligned bounding box containing the transformed
   * box */
  BoundingBox transform(Matrix m);

  /* This method returns the distance along the ray to the box if the ray
   * intersects, otherwise it returns -1.0. (The distance is 0.0 if the
   * starting position of the ray is in the box */
  Flt intersect(Ray ray);
};

#endif /* SCENE_INCLUDED */
